@@ -9,6 +9,7 @@ use std::{
 };
 
 use bch_bindgen::c;
+use log::debug;
 
 #[derive(Debug)]
 pub struct ErrnoError(pub errno::Errno);
@@ -55,14 +56,14 @@ fn handle_c_command(mut argv: Vec<String>, symlink_cmd: Option<&str>) -> i32 {
             "mkfs" => c::cmd_format(argc, argv),
             "remove-passphrase" => c::cmd_remove_passphrase(argc, argv),
             "reset-counters" => c::cmd_reset_counters(argc, argv),
-            "set-option" => c::cmd_set_option(argc, argv),
+            "set-fs-option" => c::cmd_set_option(argc, argv),
             "set-passphrase" => c::cmd_set_passphrase(argc, argv),
-            "setattr" => c::cmd_setattr(argc, argv),
+            "set-file-option" => c::cmd_setattr(argc, argv),
             "show-super" => c::cmd_show_super(argc, argv),
             "unlock" => c::cmd_unlock(argc, argv),
             "version" => c::cmd_version(argc, argv),
 
-            #[cfg(fuse)]
+            #[cfg(feature = "fuse")]
             "fusemount" => c::cmd_fusemount(argc, argv),
 
             _ => {
@@ -110,6 +111,11 @@ fn main() -> ExitCode {
         "list" => commands::list(args[1..].to_vec()).report(),
         "mount" => commands::mount(args, symlink_cmd).report(),
         "subvolume" => commands::subvolume(args[1..].to_vec()).report(),
-        _ => ExitCode::from(u8::try_from(handle_c_command(args, symlink_cmd)).unwrap()),
+        _ => {
+            let r = handle_c_command(args, symlink_cmd);
+
+            debug!("return code from C command: {r}");
+            ExitCode::from(r as u8)
+        }
     }
 }
