@@ -2,7 +2,6 @@ use std::{
     collections::HashMap,
     env,
     ffi::CString,
-    fs,
     io::{stdout, IsTerminal},
     path::{Path, PathBuf},
     ptr, str,
@@ -23,7 +22,7 @@ fn mount_inner(
     src: String,
     target: impl AsRef<std::path::Path>,
     fstype: &str,
-    mountflags: libc::c_ulong,
+    mut mountflags: libc::c_ulong,
     data: Option<String>,
 ) -> anyhow::Result<()> {
     // bind the CStrings to keep them alive
@@ -35,14 +34,32 @@ fn mount_inner(
     // convert to pointers for ffi
     let src = src.as_ptr();
     let target = target.as_ptr();
-    let data = data.map_or(ptr::null(), |data| data.as_ptr().cast());
+    let data_ptr = data.as_ref().map_or(ptr::null(), |data| data.as_ptr().cast());
     let fstype = fstype.as_ptr();
 
-    let ret = {
-        info!("mounting filesystem");
-        // REQUIRES: CAP_SYS_ADMIN
-        unsafe { libc::mount(src, target, fstype, mountflags, data) }
-    };
+    let mut ret;
+    loop {
+        ret = {
+            info!("mounting filesystem");
+            // REQUIRES: CAP_SYS_ADMIN
+            unsafe { libc::mount(src, target, fstype, mountflags, data_ptr) }
+        };
+
+        let err = errno::errno().0;
+
+        if ret == 0
+            || (err != libc::EACCES && err != libc::EROFS)
+            || (mountflags & libc::MS_RDONLY) != 0
+        {
+            break;
+        }
+
+        println!("mount: device write-protected, mounting read-only");
+        mountflags |= libc::MS_RDONLY;
+    }
+
+    drop(data);
+
     match ret {
         0 => Ok(()),
         _ => Err(crate::ErrnoError(errno::errno()).into()),
@@ -192,26 +209,114 @@ fn get_devices_by_uuid(
     Ok(get_super_blocks(uuid, &devices))
 }
 
-#[allow(clippy::type_complexity)]
-fn get_uuid_for_dev_node(
-    udev_bcachefs: &HashMap<String, Vec<String>>,
-    device: impl AsRef<Path>,
-) -> Result<(Option<Uuid>, Option<(PathBuf, bch_sb_handle)>)> {
-    let canonical = fs::canonicalize(device)?;
+fn devs_str_sbs_from_uuid(
+    udev_info: &HashMap<String, Vec<String>>,
+    uuid: &str,
+) -> anyhow::Result<(String, Vec<bch_sb_handle>)> {
+    debug!("enumerating devices with UUID {}", uuid);
 
-    if !udev_bcachefs.is_empty() {
-        let dev_node_str = canonical.into_os_string().into_string().unwrap();
+    let devs_sbs = Uuid::parse_str(uuid).map(|uuid| get_devices_by_uuid(udev_info, uuid))??;
 
-        if udev_bcachefs.contains_key(&dev_node_str) && udev_bcachefs[&dev_node_str].len() == 1 {
-            let uuid_str = udev_bcachefs[&dev_node_str][0].clone();
-            return Ok((Some(Uuid::parse_str(&uuid_str)?), None));
-        }
+    let devs_str = devs_sbs
+        .iter()
+        .map(|(dev, _)| dev.to_str().unwrap())
+        .collect::<Vec<_>>()
+        .join(":");
+
+    let sbs: Vec<bch_sb_handle> = devs_sbs.iter().map(|(_, sb)| *sb).collect();
+
+    Ok((devs_str, sbs))
+}
+
+fn devs_str_sbs_from_device(
+    udev_info: &HashMap<String, Vec<String>>,
+    device: &Path,
+) -> anyhow::Result<(String, Vec<bch_sb_handle>)> {
+    let dev_sb = read_super_silent(device)?;
+
+    if dev_sb.sb().number_of_devices() == 1 {
+        Ok((device.as_os_str().to_str().unwrap().to_string(), vec![dev_sb]))
     } else {
-        return read_super_silent(&canonical).map_or(Ok((None, None)), |sb| {
-            Ok((Some(sb.sb().uuid()), Some((canonical, sb))))
-        });
+        let uuid = dev_sb.sb().uuid();
+
+        devs_str_sbs_from_uuid(udev_info, &uuid.to_string())
+    }
+}
+
+/// If a user explicitly specifies `unlock_policy` or `passphrase_file` then use
+/// that without falling back to other mechanisms. If these options are not
+/// used, then search for the key or ask for it.
+fn handle_unlock(cli: &Cli, sb: &bch_sb_handle) -> Result<KeyHandle> {
+    if let Some(policy) = cli.unlock_policy.as_ref() {
+        return policy.apply(sb);
+    }
+
+    if let Some(path) = cli.passphrase_file.as_deref() {
+        return Passphrase::new_from_file(path).and_then(|p| KeyHandle::new(sb, &p));
+    }
+
+    let uuid = sb.sb().uuid();
+    KeyHandle::new_from_search(&uuid)
+        .or_else(|_| Passphrase::new(&uuid).and_then(|p| KeyHandle::new(sb, &p)))
+}
+
+fn cmd_mount_inner(cli: &Cli) -> Result<()> {
+    // Grab the udev information once
+    let udev_info = udev_bcachefs_info()?;
+
+    let (devices, mut sbs) =
+        if let Some(("UUID" | "OLD_BLKID_UUID", uuid)) = cli.dev.split_once('=') {
+            devs_str_sbs_from_uuid(&udev_info, uuid)?
+        } else if cli.dev.contains(':') {
+            // If the device string contains ":" we will assume the user knows the
+            // entire list. If they supply a single device it could be either the FS
+            // only has 1 device or it's only 1 of a number of devices which are
+            // part of the FS. This appears to be the case when we get called during
+            // fstab mount processing and the fstab specifies a UUID.
+
+            let sbs = cli
+                .dev
+                .split(':')
+                .map(read_super_silent)
+                .collect::<Result<Vec<_>>>()?;
+
+            (cli.dev.clone(), sbs)
+        } else {
+            devs_str_sbs_from_device(&udev_info, Path::new(&cli.dev))?
+        };
+
+    ensure!(!sbs.is_empty(), "No device(s) to mount specified");
+
+    let first_sb = &sbs[0];
+    if unsafe { bcachefs::bch2_sb_is_encrypted(first_sb.sb) } {
+        handle_unlock(cli, first_sb)?;
+    }
+
+    for sb in &mut sbs {
+        unsafe {
+            bch_bindgen::sb_io::bch2_free_super(sb);
+        }
+    }
+    drop(sbs);
+
+    if let Some(mountpoint) = cli.mountpoint.as_deref() {
+        info!(
+            "mounting with params: device: {}, target: {}, options: {}",
+            devices,
+            mountpoint.to_string_lossy(),
+            &cli.options
+        );
+
+        let (data, mountflags) = parse_mount_options(&cli.options);
+        mount_inner(devices, mountpoint, "bcachefs", mountflags, data)
+    } else {
+        info!(
+            "would mount with params: device: {}, options: {}",
+            devices, &cli.options
+        );
+
+        Ok(())
     }
-    Ok((None, None))
 }
 
 /// Mount a bcachefs filesystem by its UUID.
@@ -255,119 +360,6 @@ pub struct Cli {
     verbose: u8,
 }
 
-fn devs_str_sbs_from_uuid(
-    udev_info: &HashMap<String, Vec<String>>,
-    uuid: &str,
-) -> anyhow::Result<(String, Vec<bch_sb_handle>)> {
-    debug!("enumerating devices with UUID {}", uuid);
-
-    let devs_sbs = Uuid::parse_str(uuid).map(|uuid| get_devices_by_uuid(udev_info, uuid))??;
-
-    let devs_str = devs_sbs
-        .iter()
-        .map(|(dev, _)| dev.to_str().unwrap())
-        .collect::<Vec<_>>()
-        .join(":");
-
-    let sbs: Vec<bch_sb_handle> = devs_sbs.iter().map(|(_, sb)| *sb).collect();
-
-    Ok((devs_str, sbs))
-}
-
-fn devs_str_sbs_from_device(
-    udev_info: &HashMap<String, Vec<String>>,
-    device: impl AsRef<Path>,
-) -> anyhow::Result<(String, Vec<bch_sb_handle>)> {
-    let (uuid, sb_info) = get_uuid_for_dev_node(udev_info, device)?;
-
-    match (uuid, sb_info) {
-        (Some(uuid), Some((path, sb))) => {
-            // If we have a super block, it implies we aren't using udev db.  If we only need
-            // 1 device to mount, we'll simply return it as we're done, else we'll use the uuid
-            // to walk through all the block devices.
-            debug!(
-                "number of devices in this FS = {}",
-                sb.sb().number_of_devices()
-            );
-            if sb.sb().number_of_devices() == 1 {
-                let dev = path.into_os_string().into_string().unwrap();
-                Ok((dev, vec![sb]))
-            } else {
-                devs_str_sbs_from_uuid(udev_info, &uuid.to_string())
-            }
-        }
-        (Some(uuid), None) => devs_str_sbs_from_uuid(udev_info, &uuid.to_string()),
-        _ => Ok((String::new(), Vec::new())),
-    }
-}
-
-/// If a user explicitly specifies `unlock_policy` or `passphrase_file` then use
-/// that without falling back to other mechanisms. If these options are not
-/// used, then search for the key or ask for it.
-fn handle_unlock(cli: &Cli, sb: &bch_sb_handle) -> Result<KeyHandle> {
-    if let Some(policy) = cli.unlock_policy.as_ref() {
-        return policy.apply(sb);
-    }
-
-    if let Some(path) = cli.passphrase_file.as_deref() {
-        return Passphrase::new_from_file(path).and_then(|p| KeyHandle::new(sb, &p));
-    }
-
-    KeyHandle::new_from_search(&sb.sb().uuid())
-        .or_else(|_| Passphrase::new().and_then(|p| KeyHandle::new(sb, &p)))
-}
-
-fn cmd_mount_inner(cli: &Cli) -> Result<()> {
-    // Grab the udev information once
-    let udev_info = udev_bcachefs_info()?;
-
-    let (devices, sbs) = if let Some(("UUID" | "OLD_BLKID_UUID", uuid)) = cli.dev.split_once('=') {
-        devs_str_sbs_from_uuid(&udev_info, uuid)?
-    } else if cli.dev.contains(':') {
-        // If the device string contains ":" we will assume the user knows the
-        // entire list. If they supply a single device it could be either the FS
-        // only has 1 device or it's only 1 of a number of devices which are
-        // part of the FS. This appears to be the case when we get called during
-        // fstab mount processing and the fstab specifies a UUID.
-
-        let sbs = cli
-            .dev
-            .split(':')
-            .map(read_super_silent)
-            .collect::<Result<Vec<_>>>()?;
-
-        (cli.dev.clone(), sbs)
-    } else {
-        devs_str_sbs_from_device(&udev_info, Path::new(&cli.dev))?
-    };
-
-    ensure!(!sbs.is_empty(), "No device(s) to mount specified");
-
-    let first_sb = sbs[0];
-    if unsafe { bcachefs::bch2_sb_is_encrypted(first_sb.sb) } {
-        handle_unlock(cli, &first_sb)?;
-    }
-
-    if let Some(mountpoint) = cli.mountpoint.as_deref() {
-        info!(
-            "mounting with params: device: {}, target: {}, options: {}",
-            devices,
-            mountpoint.to_string_lossy(),
-            &cli.options
-        );
-
-        let (data, mountflags) = parse_mount_options(&cli.options);
-        mount_inner(devices, mountpoint, "bcachefs", mountflags, data)
-    } else {
-        info!(
-            "would mount with params: device: {}, options: {}",
-            devices, &cli.options
-        );
-
-        Ok(())
-    }
-}
-
 pub fn mount(mut argv: Vec<String>, symlink_cmd: Option<&str>) -> Result<()> {
     // If the bcachefs tool is being called as "bcachefs mount dev ..." (as opposed to via a
     // symlink like "/usr/sbin/mount.bcachefs dev ...", then we need to pop the 0th argument
